/*
 * ExtendedGazeteer.java
 * 
 * $Id: ExtendedGazetteer.java 2264 2011-03-28 16:35:39Z johann $
 *
 */
package at.ofai.gate.extendedgazetteer;

// NOTE: for storing lots of features without the need to fit into memory:
//   Possibly useful: Berkley DB Java Edition
//   http://www.oracle.com/technetwork/database/berkeleydb/downloads/index.html?ssSourceSiteId=ocomen
//   or this: http://code.google.com/p/orient/
//

// NOTE: VR for default gazetteer cannot be overwritten/disabled
// NOTE: deserialization of generated FSM is much slower than creating anew

// - make it possible to reuse the FSA if this PR is being loaded more than
//   onve with the same def url. Add a parameter that specifies if this should
//   be done automatically, if set true, reuse/statically save for reuse.
// - Related:


import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.util.*;

import gate.*;
import gate.creole.*;
import gate.creole.gazetteer.*;
import gate.creole.metadata.CreoleParameter;
import gate.creole.metadata.CreoleResource;
import gate.creole.metadata.RunTime;
import gate.util.*;
import java.io.BufferedReader;
import java.net.URL;

/** This is a modified version of the GATE DefaultGazetteer class. 
 * In addition to the functionality of DefaultGazetteer, 
 * this class can also provide annotations
 * for the part of a gazetteer string that directly precedes a match but follows
 * any non-word character (the prefix)
 * and the part of a gazetteer string that directly follows a match 
 * but precedes any non-word character (the suffix). That means, if the 
 * gazetteer list contains "someword" it can annotated " thisissomewordindeed "
 * so that "thisis" gets annotated as "Lookup_prefix", "someword" gets 
 * annotated as "Lookup" and "indeed" gets annotated as "Lookup_suffix".
 * <p>
 * It also gives more control over which characters mark the beginning or end
 * of a match in the document. 
 * <p>
 * Matching prefixes or suffixes only makes 
 * sense if the gazetter matches parts of words,
 * so prefix and suffix matching are deactivated if the wholeWordsOnly
 * parameter is true.
 * <p>
 * Prefix and suffix annotation can be seperately switched on and off by
 * the corresponding boolean parameters suffixAnnotations and prefixAnnotations.
 * <p>
 * Prefix and suffix annotations are created exactly as the corresponding lookup
 * annotations, but have major annotation type "Lookup_prefix" and "Lookup_suffix"
 * respectively. All features from the "Lookup" annotation are copied over
 * to any corresponding "Lookup_prefix" or "Lookup_suffix" annotation.  
 * <p>
 * Suffix annotations always include the "string" feature that contains
 * the string of the suffix. 
 * <p>
 * Lookup_prefix and Lookup  annotations include the string feature only 
 * if parameter includeStrings is set to true.)
 * <p>
 * In addition to the features provided by the GATE Default gazetteer, the
 * Lookup_prefix and Lookup annotations also include these features: 
 * <ul>
 * <li> firstcharUpper which is true if  the first letter of the corresponding 
 *   string is upper case (for Lookup_prefix and Lookup)
 * <li> atEnd which is true if a Lookup annotation matched at the end of a 
 *   word according to the current word definition (for Lookup).
 * <li>atBeginning which is true if a Lookup annotation matched at the 
 *   beginning of a word according to the current word definition.
 * </ul>
 * <p>
 * All parameters except encoding, gazetteerFeatureSeparator, and listsURL,
 * are defined to be runtime parameters so it is 
 * much easier to change them during debugging without the need to re-create
 * the processing resource.
 * <p>
 * The following parameters influence whether characters before or after
 * a matching string in the document are regarded as word separators or not.
 * These parameters do not influence how any of the characters within the
 * string that occurs in the gazetteer list are handled: these characters
 * always have to match exactly as they occur in the gazetteer list.
 * parameters: 
 * <ul>
 * <li> wordCharsClass: 
 *   <ul>
 *    <li>NONWHITESPACE: gazetteer strings are delimited by whitespace characters. If
 *      matching with wholWordsOnly=true, only strings that are preceded and
 *      followed by whitespace characters are considered "whole words". Note that
 *      whitespace that is part of the gazetteer list entry is still part
 *      of the string that must be matched successfully.
 *    <li>LETTERS: words only consist of letters (unicode class). Everything
 *        else (including digits or special characters) is interpreted 
 *        as word boundary. 
 *    <li>DIGITS: words only consist of digits (unicode class). Everything
 *        else (including letters and special characters) is interpreted
 *        as word boundary
 *    <li>LETTERSORDIGITS: words consist of digits or letters
 *   </ul>
 * 
 *  <li>wordChars: a string made up of additional characters that should be 
 *    accepted for words. Whitespace will be removed.
 *    You might want to add e.g. a hyphen here.
 *  <li>wordBoundaryChars: a string made up of additional characters that 
 *    should be interpreted as word boundaries. Whitespace will be removed. 
 *  </ul>
 *
 *  Whitespace will ALWAYS be interpreted as word boundary, combining spacing
 *  mark and non spacing mark will always be interpreted as part of a word. 
 *  <p>
 *  The word characters as defined here will only influence how the characters
 *  *outside* the actual gazetteer match will be processed, i.e. how 
 *  suffixes and prefixes are found. That means that an entry in a gazetteer
 *  list can contain non-word characters and still match, e.g. 
 *  "worda wordb" will match "theworda wordbs" even though the space is a 
 *  non-word character and will generate Lookup_prefix.string = "the" and 
 *  Lookup_suffix.string = "s". 
 *  <p>
 *  NOTE: features like "firstcharUpper" are set to "true" or "false" as
 *  strings, not booleans.
 *  
 *  @author Valentin Tablan
 *  @author Borislav Popov
 *  @author Johann Petrak
 */
@CreoleResource(name = "ExtendedGazetteerCompat",
comment = "An extension of the GATE DefaultGazetteer that can handle prefixes and suffixes")
public class ExtendedGazetteer
  extends DefaultGazetteer {

  private static final long serialVersionUID = 1122334411234578432L;
;
  private Boolean prefixAnnotations = false;
  private Boolean suffixAnnotations = false;
  private Boolean includeStrings = false;
  private WCClass wordCharsClass = WCClass.NONWHITESPACE;
  private String wordBoundaryChars = "";
  private String wordChars = "";
  static boolean isShownVersion = false;

  public Boolean getPrefixAnnotations() {
    return prefixAnnotations;
  }

  @CreoleParameter(comment = "Generate prefix annotations?", defaultValue="true")
  @RunTime
  public void setPrefixAnnotations(Boolean newPrefixAnnotations) {
    prefixAnnotations = newPrefixAnnotations;
  }

  public Boolean getSuffixAnnotations() {
    return suffixAnnotations;
  }

  @CreoleParameter(comment = "Generate suffix annotations?", defaultValue="true")
  @RunTime
  public void setSuffixAnnotations(Boolean newSuffixAnnotations) {
    suffixAnnotations = newSuffixAnnotations;
  }

  public Boolean getIncludeStrings() {
    return includeStrings;
  }

  @CreoleParameter(comment = "Include original string as feature?", defaultValue="false")
  @RunTime
  public void setIncludeStrings(Boolean newIncludeStrings) {
    includeStrings = newIncludeStrings;
  }

  public WCClass getWordCharsClass() {
    return wordCharsClass;
  }


  @CreoleParameter(comment = "Types of characters that make up words?", defaultValue="NONWHITESPACE")
  @RunTime
  public void setWordCharsClass(WCClass newWordCharsClass) {
    wordCharsClass = newWordCharsClass;
  }

  public String getWordBoundaryChars() {
    return wordBoundaryChars;
  }

  @CreoleParameter(comment = "Additional word boundary characters?", defaultValue="")
  @RunTime
  public void setWordBoundaryChars(String newWordBoundaryChars) {
    wordBoundaryChars = newWordBoundaryChars;
  }

  public String getWordChars() {
    return wordChars;
  }

  @CreoleParameter(comment = "Additional word characters?", defaultValue="-")
  @RunTime
  public void setWordChars(String newWordChars) {
    wordChars = newWordChars;
  }

  @CreoleParameter(comment = "Non-editable and memory saving mode, also enables FSM re-use for multiple copies", defaultValue="true")
  public void setMemorySavingMode(Boolean yesno) {
    memorySavingMode = yesno;
  }
  public Boolean getMemorySavingMode() {
    return memorySavingMode;
  }
  protected Boolean memorySavingMode;


  protected String unescapedSeparator = null;

  protected static Map<URL,FSM> loadedGazetteers = new HashMap<URL,FSM>();

  protected class FSM {
    public Set<FSMState> fsmStates;
    public FSMState initialState;
    public FSM(Set<FSMState> states, FSMState init) {
      fsmStates = states;
      initialState = init;
    }
    public int refcount = 0;
  }

  /** 
   * Build a gazetter using the default lists from the gate resources
   */
  public ExtendedGazetteer() {
  }

  public Resource init() throws ResourceInstantiationException {
    /* do the init stuff we need specially for this subclass */
    if (!isShownVersion) {
      System.out.println("Plugin version: $Id: ExtendedGazetteer.java 2264 2011-03-28 16:35:39Z johann $");
      isShownVersion = true;
    }

    // silently remove all whitespace from the wordChars and also from 
    // wordBoundaryChars (because Whitespace will always be a word Boundary)
    wordChars.replaceAll("\\p{Z}", "");
    wordBoundaryChars.replaceAll("\\p{Z}", "");
    if(memorySavingMode) {
      FSM fsm = loadedGazetteers.get(listsURL);
      if(fsm != null) {
        fsmStates = fsm.fsmStates;
        initialState = fsm.initialState;
        fsm.refcount++;
        System.out.println("Reusing already generated FSM");
      } else {
        this.fsmStates = new HashSet();
        initialState = new FSMState(this);
        if (gazetteerFeatureSeparator != null && !gazetteerFeatureSeparator.isEmpty()) {
          unescapedSeparator = Strings.unescape(gazetteerFeatureSeparator);
        }
        try {
          loadData();
        } catch (Exception ex) {
          throw new ResourceInstantiationException("Could not initialize", ex);
        }
        fsm = new FSM(fsmStates,initialState);
        fsm.refcount++;
        loadedGazetteers.put(listsURL, fsm);
      }
      return this;
    } else {
      return super.init();
    }
  }

  public void cleanup() {
    FSM fsm = loadedGazetteers.get(listsURL);
    fsm.refcount--;
    if(fsm.refcount == 0) {
      loadedGazetteers.remove(listsURL);
      System.out.println("Removing FSM");
    }
  }

  protected void loadData() throws UnsupportedEncodingException, IOException, ResourceInstantiationException  {
    BufferedReader defReader =
      new BomStrippingInputStreamReader((listsURL).openStream(), encoding);
    String line;
    while (null != (line = defReader.readLine())) {
       String[] fields = line.split(":");
       if(fields.length == 0) {
         System.err.println("Empty line in file "+listsURL);
       } else {
         String listFileName = "";
         String majorType = "";
         String minorType = "";
         String languages = "";
         listFileName = fields[0];
         if(fields.length > 1) {
           majorType = fields[1];
         }
         if(fields.length > 2) {
           minorType = fields[2];
         }
         if(fields.length > 3) {
           languages = fields[3];
         }
         if(fields.length > 4) {
           defReader.close();
           throw new ResourceInstantiationException("Line has more that 4 fields in def file "+listsURL);
         }
         System.out.println("Reading from "+listFileName+", "+majorType+"/"+minorType+"/"+languages);
         loadListFile(listFileName,majorType,minorType,languages);
      }
    } //while
    defReader.close();
  }

  void loadListFile(String listFileName,String majorType,String minorType,String languages) throws MalformedURLException, IOException {
    // create a completely new fsm
    // query SOLR
    Lookup defaultLookup = new Lookup(listFileName, majorType, minorType, languages);
    URL lurl = new URL(listsURL,listFileName);
    BufferedReader listReader =
      new BomStrippingInputStreamReader(lurl.openStream(), encoding);
    String line;
    int lines = 0;
    while (null != (line = listReader.readLine())) {
      GazetteerNode node = new GazetteerNode(line, unescapedSeparator, false);
      Lookup lookup = defaultLookup;
      Map<String,String> fm = node.getFeatureMap();
      if(fm != null && fm.size() > 0) {
        lookup = new Lookup(listFileName, majorType, minorType, languages);
        // if the only keys we have in the map are one of majorType, minorType,
        // or language, then just overwrite the values in the new lookup
        // object, we wont need an extra new feature map for that!
        Set<String> keyset = fm.keySet();
        if(keyset.size() <= 3) {
          Map<String,String> newfm = null;
          for(String key : keyset) {
            if(key.equals("majorType")) {
              lookup.majorType = fm.get("majorType");
            } else if(key.equals("minorType")) {
              lookup.minorType = fm.get("minorType");
            } else if(key.equals("language")) {
              lookup.languages = fm.get("language");
            } else {
              // some other key, we copy over to the new map
              if(newfm == null) {
                newfm = new HashMap<String,String>();
              }
              newfm.put(key, fm.get(key));
            }
          }
          // if we had to allocate the map, store it
          if(newfm != null) {
            lookup.features = newfm;
          }
        } else {
          lookup.features = fm;
        }
      }
      addLookup(node.getEntry(), lookup);
      lines++;
    } // while
    System.out.println("Lines read: "+lines);
  }



  /**
   * Tests whether a character is internal to a word (i.e. if it's a letter or
   * a combining mark (spacing or not)).
   * @param ch the character to be tested
   * @return a boolean value
   */
  public boolean isWithinWord(char ch) {
    if (Character.getType(ch) == Character.COMBINING_SPACING_MARK
      || Character.getType(ch) == Character.NON_SPACING_MARK) {
      return true;
    }
    if (Character.isWhitespace(ch)) {
      return false;
    }

    if (wordChars.indexOf(ch) != -1) {
      return true;
    }
    if (wordBoundaryChars.indexOf(ch) != -1) {
      return false;
    }

    if (wordCharsClass == WCClass.LETTERS) {
      return Character.isLetter(ch);
    } else if (wordCharsClass == WCClass.DIGITS) {
      return Character.isDigit(ch);
    } else if (wordCharsClass == WCClass.LETTERSORDIGITS) {
      return (Character.isDigit(ch) || Character.isLetter(ch));
    }
    return true;   // in case we have the default WC_NONWHITESPACE
  }

  /**
   * This method runs the gazetteer. It assumes that all the needed parameters
   * are set. If they are not, an exception will be fired.
   * 
   * @throws ExecutionException 
   */
  // This is a copy of the original DefaultGazetteer class with
  // the changes to support the additional functions.
  @Override
  public void execute() throws ExecutionException {
    interrupted = false;
    AnnotationSet annotationSet;
    //check the input
    if (document == null) {
      throw new ExecutionException(
        "No document to process!");
    }

    if (annotationSetName == null
      || annotationSetName.equals("")) {
      annotationSet = document.getAnnotations();
    } else {
      annotationSet = document.getAnnotations(annotationSetName);
    }

    fireStatusChanged("Performing look-up in " + document.getName() + "...");
    int original_length = document.getContent().toString().length();
    // We are working on a copy of the string, so we will append whitespace to
    // avoid unnecessary complex end of document behavior during matching.
    String content = document.getContent().toString() + "          ";
    int length = content.length();

    char currentChar;

    FSMState currentState = initialState;
    FSMState nextState;
    FSMState lastMatchingState = null;
    int matchedRegionEnd = 0;
    int matchedRegionStart = 0;
    int charIdx = 0;

    int oldCharIdx = 0;
    FeatureMap fm;
    FeatureMap fmprefix;
    FeatureMap fmsuffix;
    Lookup currentLookup;

    int lastNonWordIdx = -1;
    int lastLastNonWordIdx = -1;
    boolean doPrefixAnnotation = getPrefixAnnotations() && !wholeWordsOnly.booleanValue();
    boolean doSuffixAnnotation = getSuffixAnnotations() && !wholeWordsOnly.booleanValue();
    boolean matchedAtEnd;
    boolean matchedAtBeginning;
    int suffixFrom;
    int suffixTo;
    int prefixFrom;
    int prefixTo;
    String suffixString;

    while (charIdx < length) {
      currentChar = content.charAt(charIdx);
      if (Character.isWhitespace(currentChar)) {
        currentChar = ' ';
      } else {
        currentChar = caseSensitive.booleanValue()
          ? currentChar
          : Character.toUpperCase(currentChar);
      }
      if (!isWithinWord(currentChar)) {
        lastLastNonWordIdx = lastNonWordIdx;
        lastNonWordIdx = charIdx;
        // if the gazetteer list contains "asd fgh" it will always match
        // "asd    fgh" too: whitespace within the string is being compressed.
        // However if we e.g. treat everything except digits as word boundaries,
        // "123x456" will not match with "123xxxxxx456".
        // We could set the currentChar to " " here but the FSM would still
        // have to know that "x" has to be treated as whitespace here, which
        // it does not and which we cannot do without cloning the whole FSM
        // implementation.
      }
      nextState = currentState.next(currentChar);
      if (nextState == null) {
        //the matching stopped

        //if we had a successful match then act on it;
        if (lastMatchingState != null) {
          //let's add the new annotation(s)
          Iterator lookupIter = lastMatchingState.getLookupSet().iterator();

          // if we have a match and now are at a whitespace, we want the previous
          // one: 
          if (!isWithinWord(currentChar)) {
            lastNonWordIdx = lastLastNonWordIdx;
          }
          while (lookupIter.hasNext()) {
            currentLookup = (Lookup) lookupIter.next();
            fm = Factory.newFeatureMap();
            fm.put(LOOKUP_MAJOR_TYPE_FEATURE_NAME, currentLookup.majorType);
            if (null != currentLookup.oClass && null != currentLookup.ontology) {
              fm.put(LOOKUP_CLASS_FEATURE_NAME, currentLookup.oClass);
              fm.put(LOOKUP_ONTOLOGY_FEATURE_NAME, currentLookup.ontology);
            }
            if (null != currentLookup.minorType) {
              fm.put(LOOKUP_MINOR_TYPE_FEATURE_NAME, currentLookup.minorType);
              if (null != currentLookup.languages) {
                fm.put("language", currentLookup.languages);
              }
            }
            fm.put("firstcharCategory", Character.getType(content.charAt(matchedRegionStart)));
            if (Character.isUpperCase(content.charAt(matchedRegionStart))) {
              fm.put("firstcharUpper", true);
            } else {
              fm.put("firstcharUpper", false);
            }
            if (!isWithinWord(content.charAt(Math.min(content.length() - 1, matchedRegionEnd + 1)))) {
              fm.put("atEnd", true);
              matchedAtEnd = true;
            } else {
              fm.put("atEnd", false);
              matchedAtEnd = false;
            }
            if (matchedRegionStart > 0 && isWithinWord(content.charAt(matchedRegionStart - 1))) {
              fm.put("atBeginning", false);
              matchedAtBeginning = false;
            } else {
              fm.put("atBeginning", true);
              matchedAtBeginning = true;
            }
            if (includeStrings) {
              fm.put("string", content.substring(matchedRegionStart, matchedRegionEnd + 1));
            }
            if (null != currentLookup.features) {
              fm.putAll(currentLookup.features);
            }

            //try {
            try {
              annotationSet.add(new Long(matchedRegionStart),
                new Long(matchedRegionEnd + 1),
                LOOKUP_ANNOTATION_TYPE,
                fm);
            } catch (InvalidOffsetException ex) {
              throw new GateRuntimeException("Invalid offset exception - doclen/from/to="
                + original_length + "/" + matchedRegionStart + "/" + (matchedRegionEnd + 1) + " / " + ex);
            }
            // This was originally:
            //  if(doPrefixAnnotation && (matchedRegionStart>(lastNonWordIdx+1)))
            // Using the lastNonWordIdx for prefix checking does not work
            // if a character in the gazetteer word is not a word character
            // but we still want to find a prefix, e.g.
            // the gazetteer word is "word/word2" and we want to find
            // "the" in "theword/word2".
            // For this we have now added the matchedAtBeginning flag similar
            // to matchedAtEnd and if we are not at the beginning, we now
            // search back from the location before the start of the match
            // until we find a nonword character.
            if (doPrefixAnnotation && !matchedAtBeginning) {
              prefixTo = matchedRegionStart; // if we get here, this should exist
              prefixFrom = prefixTo - 1;
              while ((prefixFrom - 1) > 0 && isWithinWord(content.charAt(prefixFrom - 1))) {
                prefixFrom--;
              }
              fmprefix = Factory.newFeatureMap();
              fmprefix.put(LOOKUP_MAJOR_TYPE_FEATURE_NAME, currentLookup.majorType);
              if (null != currentLookup.oClass && null != currentLookup.ontology) {
                fmprefix.put(LOOKUP_CLASS_FEATURE_NAME, currentLookup.oClass);
                fmprefix.put(LOOKUP_ONTOLOGY_FEATURE_NAME, currentLookup.ontology);
              }
              if (null != currentLookup.minorType) {
                fmprefix.put(LOOKUP_MINOR_TYPE_FEATURE_NAME, currentLookup.minorType);
                if (null != currentLookup.languages) {
                  fmprefix.put("language", currentLookup.languages);
                }
              }
              fmprefix.put("firstcharCategory", Character.getType(content.charAt(prefixFrom)));
              if (Character.isUpperCase(content.charAt(prefixFrom))) {
                fmprefix.put("firstcharUpper", true);
              } else {
                fmprefix.put("firstcharUpper", false);
              }
              if (includeStrings) {
                fmprefix.put("string", content.substring(prefixFrom, prefixTo));
              }
              if (null != currentLookup.features) {
                fmprefix.putAll(currentLookup.features);
              }
              try {
                annotationSet.add(new Long(prefixFrom),
                  new Long(prefixTo),
                  "Lookup_prefix",
                  fmprefix);
              } catch (InvalidOffsetException ex) {
                throw new GateRuntimeException("Invalid offset exception - doclen/from/to="
                  + original_length + "/" + prefixFrom + "/" + (prefixTo) + " / " + ex);
              }
            }
            if (doSuffixAnnotation && !matchedAtEnd) {
              // get the suffix string start and end offsets, if there really
              // is one

              suffixFrom = Math.min(content.length() - 1, matchedRegionEnd + 1);
              suffixTo = suffixFrom;
              while (suffixTo < content.length() && isWithinWord(content.charAt(suffixTo))) {
                suffixTo++;
              }
              if (suffixTo > suffixFrom) {
                suffixString = content.substring(suffixFrom, suffixTo);
                fmsuffix = Factory.newFeatureMap();
                fmsuffix.put(LOOKUP_MAJOR_TYPE_FEATURE_NAME, currentLookup.majorType);
                if (null != currentLookup.oClass && null != currentLookup.ontology) {
                  fmsuffix.put(LOOKUP_CLASS_FEATURE_NAME, currentLookup.oClass);
                  fmsuffix.put(LOOKUP_ONTOLOGY_FEATURE_NAME, currentLookup.ontology);
                }
                if (null != currentLookup.minorType) {
                  fmsuffix.put(LOOKUP_MINOR_TYPE_FEATURE_NAME, currentLookup.minorType);
                  if (null != currentLookup.languages) {
                    fmsuffix.put("language", currentLookup.languages);
                  }
                }
                fmsuffix.put("string", suffixString);
                if (null != currentLookup.features) {
                  fmsuffix.putAll(currentLookup.features);
                }
                try {
                  annotationSet.add(new Long(suffixFrom),
                    new Long(suffixTo),
                    "Lookup_suffix",
                    fmsuffix);
                } catch (InvalidOffsetException ex) {
                  throw new GateRuntimeException("Invalid offset exception - doclen/from/to="
                    + original_length + "/" + suffixFrom + "/" + suffixTo + " / " + ex);
                }
              }
            }
            //} catch(InvalidOffsetException ioe) {
            //  throw new LuckyException(ioe.toString());
            //}
          }//while(lookupIter.hasNext())
          lastMatchingState = null;
        } // if(lastMatchingState != null) // end we had a successful match

        //reset the FSM
        charIdx = matchedRegionStart + 1;
        matchedRegionStart = charIdx;
        currentState = initialState;

      } else {  // if(nextState==null) i.e. the matching goes on ....
        currentState = nextState;
        //if we have a successful state then store it
        if (currentState.isFinal()
          && ((!wholeWordsOnly.booleanValue())
          || ( // wholeWordsOnly
          (matchedRegionStart == 0
          || !isWithinWord(content.charAt(matchedRegionStart - 1)))
          && (charIdx + 1 >= content.length()
          || !isWithinWord(content.charAt(charIdx + 1)))))) {
          matchedRegionEnd = charIdx;
          lastMatchingState = currentState;
        } // end if (currentState.isFinal() && ....)
        charIdx++;
        if (charIdx == content.length()) {
          //we can't go on, use the last matching state and restart matching
          //from the next char
          if (lastMatchingState != null) {
            //let's add the new annotation(s)
            Iterator lookupIter = lastMatchingState.getLookupSet().iterator();

            while (lookupIter.hasNext()) {
              currentLookup = (Lookup) lookupIter.next();
              fm = Factory.newFeatureMap();
              fm.put(LOOKUP_MAJOR_TYPE_FEATURE_NAME, currentLookup.majorType);
              if (null != currentLookup.oClass && null != currentLookup.ontology) {
                fm.put(LOOKUP_CLASS_FEATURE_NAME, currentLookup.oClass);
                fm.put(LOOKUP_ONTOLOGY_FEATURE_NAME, currentLookup.ontology);
              }
              if (null != currentLookup.minorType) {
                fm.put(LOOKUP_MINOR_TYPE_FEATURE_NAME, currentLookup.minorType);
                if (null != currentLookup.languages) {
                  fm.put("language", currentLookup.languages);
                }
              }
              // try {
              //System.out.println("Found a match2:"+lastNonWordIdx+"/"+matchedRegionStart);
              if (doPrefixAnnotation && (matchedRegionStart > (lastNonWordIdx + 1))) {
                //System.out.println("Have a prefix match2!");
                try {
                  annotationSet.add(new Long(lastNonWordIdx + 1),
                    new Long(matchedRegionStart),
                    "Lookup_prefix",
                    fm);
                } catch (InvalidOffsetException ex) {
                  throw new GateRuntimeException("Invalid offset exception - doclen/from/to="
                    + original_length + "/" + (lastNonWordIdx + 1) + "/" + (matchedRegionStart) + " / " + ex);
                }
                // TODO: add string of prefix to annotation, add indicator
                // if first character is upper case
              }
              try {
                annotationSet.add(new Long(matchedRegionStart),
                  new Long(matchedRegionEnd + 1),
                  LOOKUP_ANNOTATION_TYPE,
                  fm);
              } catch (InvalidOffsetException ex) {
                throw new GateRuntimeException("Invalid offset exception - doclen/from/to="
                  + original_length + "/" + matchedRegionStart + "/" + (matchedRegionEnd + 1) + " / " + ex);
              }
              //} catch(InvalidOffsetException ioe) {
              //  throw new LuckyException(ioe.toString());
              //}
            }//while(lookupIter.hasNext())
            lastMatchingState = null;
          }

          //reset the FSM
          charIdx = matchedRegionStart + 1;
          matchedRegionStart = charIdx;
          currentState = initialState;
        }
      } // end of if(nextState==null) 

      // what follows is done in both cases: matching stopped or goes on
      if (charIdx - oldCharIdx > 256) {
        fireProgressChanged((100 * charIdx) / length);
        oldCharIdx = charIdx;
        if (isInterrupted()) {
          throw new ExecutionInterruptedException(
            "The execution of the " + getName()
            + " gazetteer has been abruptly interrupted!");
        }
      }
    } // while(charIdx < length)

    // After we matched the whole document: we might have just stopped with 
    // a match. In the original DefaultGazetteer, there was an additional
    // code block to process these matches, but we have deleted this here.
    // Instead, we use the following trick: before matching starts, the
    // content gets 10 blanks appended, so, assuming that no hazetteer word
    // will end in 10 blanks, a match at the end will never occur. 
    fireProcessFinished();
    fireStatusChanged("Lookup complete!");
  } // execute
} // ExtendedGazetteer

